import torch
import pytorch_lightning as pl
from torch import nn
from typing import Any
from functools import partial
from torch.optim.lr_scheduler import LambdaLR
from pytorch_lightning.utilities import rank_zero_info
from .env import BaseEnv


class BaseModel(pl.LightningModule):
    def __init__(
        self,
        env: BaseEnv,
        model: nn.Module,
        lr_scheduler: str = "cosine-decay",
        learning_rate: float = 2e-4,
        weight_decay: float = 1e-4,
    ):
        super(BaseModel, self).__init__()
        self.env = env
        self.model = model
        self.lr_scheduler = lr_scheduler
        self.num_training_steps_cached = None
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay

    def load_data(self):
        self.env.load_data()

    def train_dataloader(self):
        return self.env.train_dataloader()

    def val_dataloader(self):
        return self.env.val_dataloader()

    def test_dataloader(self):
        return self.env.test_dataloader()

    def get_net(self, network_type: str):
        self.network_type = network_type
        return NotImplementedError

    def configure_optimizers(self):
        """ """
        rank_zero_info(
            "Parameters: %d" % sum([p.numel() for p in self.model.parameters()])
        )
        rank_zero_info("Training steps: %d" % self.get_total_num_training_steps())

        if self.lr_scheduler == "constant":
            return torch.optim.AdamW(
                self.model.parameters(),
                lr=self.learning_rate,
                weight_decay=self.weight_decay,
            )
        else:
            optimizer = torch.optim.AdamW(
                self.model.parameters(),
                lr=self.learning_rate,
                weight_decay=self.weight_decay,
            )
            scheduler = get_schedule_fn(
                self.lr_scheduler, self.get_total_num_training_steps()
            )(optimizer)

        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "interval": "step",
            },
        }

    def get_total_num_training_steps(self) -> int:
        """
        Total training steps inferred from datamodule and devices.
        """
        if self.num_training_steps_cached is not None:
            return self.num_training_steps_cached

        dataset = self.train_dataloader()
        if self.trainer.max_steps and self.trainer.max_steps > 0:
            return self.trainer.max_steps

        dataset_size = (
            self.trainer.limit_train_batches * len(dataset)
            if self.trainer.limit_train_batches != 0
            else len(dataset)
        )
        num_devices = max(1, self.trainer.num_devices)
        effective_batch_size = self.trainer.accumulate_grad_batches * num_devices
        self.num_training_steps_cached = (
            dataset_size // effective_batch_size
        ) * self.trainer.max_epochs

        return self.num_training_steps_cached

    def shared_step(self, batch: Any, batch_idx: int, phase: str):
        """Shared step between train/val/test. To be implemented in subclasses."""
        raise NotImplementedError(
            "Shared step is required to implemented in subclasses."
        )

    def training_step(self, batch: Any, batch_idx: int):
        # To use new data every epoch, we need to call reload_dataloaders_every_epoch=True in Trainer
        return self.shared_step(batch, batch_idx, phase="train")

    def validation_step(self, batch: Any, batch_idx: int):
        return self.shared_step(batch, batch_idx, phase="val")

    def test_step(self, batch: Any, batch_idx: int):
        return self.shared_step(batch, batch_idx, phase="test")

    def forward(self, x, graph, edge_index, t=None) -> torch.Tensor:
        return self.model(x, graph=graph, edge_index=edge_index, timesteps=t)

    def solve(self, data: Any, batch_size: int = 16, device="cpu"):
        """solve function, return heatmap"""
        raise NotImplementedError("solve is required to implemented in subclasses.")

    def load_weights(self):
        """load state dict from checkpoint"""
        raise NotImplementedError(
            "``load_ckpt`` is required to implemented in subclasses."
        )


def get_schedule_fn(scheduler, num_training_steps):
    """Returns a callable scheduler_fn(optimizer).
    Todo: Sanitize and unify these schedulers...
    """

    def get_one_cycle(optimizer, num_training_steps):
        """Simple single-cycle scheduler. Not including paper/fastai three-phase things or asymmetry."""

        def lr_lambda(current_step):
            if current_step < num_training_steps / 2:
                return float(current_step / (num_training_steps / 2))
            else:
                return float(2 - current_step / (num_training_steps / 2))

        return LambdaLR(optimizer, lr_lambda, -1)

    if scheduler == "cosine-decay":
        scheduler_fn = partial(
            torch.optim.lr_scheduler.CosineAnnealingLR,
            T_max=num_training_steps,
            eta_min=0.0,
        )
    elif scheduler == "one-cycle":  # this is a simplified one-cycle
        scheduler_fn = partial(
            get_one_cycle,
            num_training_steps=num_training_steps,
        )
    else:
        raise ValueError(f"Invalid schedule {scheduler} given.")
    return scheduler_fn



""" Concorde base layer.

Cython wrappers around Concorde API.

"""
import numpy as np
cimport numpy as np

cdef extern from "concorde.h":
    struct CCdatagroup:
        double *x
        double *y
        double *z

    struct CCrandstate:
        pass

    int CCutil_gettsplib(char *datname, int *ncount, CCdatagroup *dat)

    int CCtsp_solve_dat (int ncount, CCdatagroup *indat, int *in_tour,
        int *out_tour, double *in_val, double *optval, int *success,
        int *foundtour, char *name, double *timebound, int *hit_timebound,
        int silent, CCrandstate *rstate)

    double CCutil_real_zeit()
    void CCutil_sprand (int seed, CCrandstate *r)

    void CCutil_freedatagroup (CCdatagroup *dat)


cdef class _CCdatagroup:

    cdef CCdatagroup c_data
    cdef bint initialized
    cdef int ncount

    def __cinit__(self):
        self.initialized = False

    def __dealloc__(self):
        if self.initialized:
            CCutil_freedatagroup(&(self.c_data))

    @property
    def x(self):
        cdef double[:] x_data
        if self.initialized:
            x_data = <double[:self.ncount]>self.c_data.x
            return np.asarray(x_data)
        else:
            return np.array([])

    @property
    def y(self):
        cdef double[:] y_data
        if self.initialized:
            y_data = <double[:self.ncount]>self.c_data.y
            return np.asarray(y_data)
        else:
            return np.array([])

    @property
    def z(self):
        cdef double[:] y_data
        if self.initialized:
            z_data = <double[:self.ncount]>self.c_data.z
            return np.asarray(z_data)
        else:
            return np.array([])


def _CCutil_gettsplib(str fname):
    cdef int ncount, retval
    cdef _CCdatagroup dat

    dat = _CCdatagroup()

    retval = CCutil_gettsplib(fname.encode('utf-8'), &ncount, &dat.c_data)
    if retval == 0:
        dat.initialized = True
        dat.ncount = ncount
        return ncount, dat
    else:
        return -1, None


def _CCtsp_solve_dat(
        int ncount, _CCdatagroup ingroup,
        str name, double timebound, int silent, int seed=0):

    cdef:
        int *in_tour = NULL
        double *in_val = NULL      # initial upper bound
        double opt_val = 0         # value of the optimal tour
        int success = 0            # set to 1 if the run finishes normally
        int foundtour = 0          # set to 1 if a tour has been found
        double *_timebound = NULL  # NULL if no timebound, >= 0 otherwise
        int hit_timebound = 0
        int retval

        # Random state used by the solver
        CCrandstate rstate

        # Output tour
        np.ndarray[int, ndim=1] out_tour

    out_tour = np.zeros(ncount, dtype=np.int32)

    if seed != 0:
        seed = <int>CCutil_real_zeit()
    CCutil_sprand (seed, &rstate)

    if timebound > 0:
        _timebound = &timebound

    retval = CCtsp_solve_dat(ncount, &ingroup.c_data, in_tour, &out_tour[0],
                             in_val, &opt_val, &success, &foundtour,
                             name.encode('utf-8'), _timebound, &hit_timebound,
                             silent, &rstate)

    return out_tour, opt_val, bool(success), bool(foundtour), bool(hit_timebound)



import torch
import torch.utils.data
from .data import TSPDataset
from ml4co_kit.learning.env import BaseEnv
from torch_geometric.data import DataLoader as GraphDataLoader


class TSPEnv(BaseEnv):
    def __init__(
        self,
        nodes_num: int = None,
        mode: str = None,
        train_path: str = None,
        val_path: str = None,
        test_path: str = None,
        train_batch_size: int = 1,
        val_batch_size: int = 1,
        test_batch_size: int = 1,
        valid_samples: int = 1280,
        num_workers: int = 1,
    ):
        super().__init__(
            name="tsp",
            mode=mode,
            train_path=train_path,
            val_path=val_path,
            test_path=test_path,
            train_batch_size=train_batch_size,
            val_batch_size=val_batch_size,
            test_batch_size=test_batch_size,
            num_workers=num_workers,
        )
        self.nodes_num = nodes_num
        self.valid_samples = valid_samples
        self.load_data()

    def load_data(self):
        if self.mode == "train":
            self.train_dataset = TSPDataset(
                data_path=self.train_path,
                mode="train",
            )
            self.val_dataset = TSPDataset(
                data_path=self.val_path,
                mode="val",
            )
        elif self.mode == "test":
            self.test_dataset = TSPDataset(
                data_path=self.test_path,
                mode="test",
            )
        else:
            pass

    def train_dataloader(self):
        train_dataloader = GraphDataLoader(
            self.train_dataset,
            batch_size=self.train_batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=True,
            drop_last=True,
        )
        return train_dataloader

    def val_dataloader(self):
        val_dataset = torch.utils.data.Subset(
            dataset=self.val_dataset, indices=range(self.valid_samples)
        )
        val_dataloader = GraphDataLoader(
            val_dataset, batch_size=self.val_batch_size, shuffle=False
        )
        return val_dataloader

    def test_dataloader(self):
        # force the test batch size to be 1
        self.test_batch_size = 1
        test_dataloader = GraphDataLoader(
            self.test_dataset,
            batch_size=self.test_batch_size,
        )
        return test_dataloader
